@@ -10,7 +10,10 @@ class ListsPluginConfig(PluginConfig):
     author = "Devon Mar"
     base_url = "lists"
     required_settings = []
-    default_settings = {}
+    default_settings = {
+        "as_cidr": True,
+        "service_primary_ips": True
+    }
 
 
 config = ListsPluginConfig
