@@ -1,7 +1,9 @@
 from rest_framework import routers
+from django.urls import path
 from .views import (
-    AggregateListViewSet, DevicesListViewSet, IPAddressListViewSet, ListsRootView, PrefixListViewSet,
-    PrometheusDeviceSD, PrometheusVirtualMachineSD, ServiceListviewSet, TagsListViewSet, VirtualMachinesListViewSet
+    AggregateListViewSet, DevicesListViewSet, DevicesVMsListView, IPAddressListViewSet, ListsRootView,
+    PrefixListViewSet, PrometheusDeviceSD, PrometheusVirtualMachineSD, ServiceListviewSet, TagsListViewSet,
+    VirtualMachinesListViewSet
 )
 
 router = routers.DefaultRouter()
@@ -16,4 +18,6 @@ router.register("devices", DevicesListViewSet, basename="devices")
 router.register("virtual-machines", VirtualMachinesListViewSet, basename="virtual-machines")
 router.register("prometheus-devices", PrometheusDeviceSD)
 router.register("prometheus-vms", PrometheusVirtualMachineSD)
-urlpatterns = router.urls
+
+urlpatterns = [path("devices-vms/", DevicesVMsListView.as_view())]
+urlpatterns += router.urls
