@@ -1,17 +1,20 @@
+from netbox_lists.api.filtersets import CustomPrefixFilterSet
 import operator
 from functools import reduce
+from typing import List, Union
 
 from django.db.models import Q
 from rest_framework import status
 from rest_framework.generics import get_object_or_404
 from rest_framework.response import Response
 from rest_framework.routers import APIRootView
+from rest_framework.views import APIView
 from rest_framework.viewsets import GenericViewSet
 from rest_framework.renderers import JSONRenderer, BrowsableAPIRenderer
 from rest_framework.request import Request
 
 from ipam.models import Prefix, Aggregate, IPAddress, Service
-from ipam.filtersets import PrefixFilterSet, IPAddressFilterSet, AggregateFilterSet, ServiceFilterSet
+from ipam.filtersets import IPAddressFilterSet, AggregateFilterSet, ServiceFilterSet
 from dcim.models import Device
 from dcim.filtersets import DeviceFilterSet
 from virtualization.models import VirtualMachine
@@ -22,18 +25,22 @@ from drf_yasg import openapi
 from drf_yasg.utils import swagger_auto_schema
 
 from .renderers import PlainTextRenderer
-from .utils import as_cidr, devicesVmPrimaryList, validateFamilyParam
+from .utils import (
+    device_vm_primary_list, format_ipn, get_as_cidr,
+    get_family_param, get_service_ips, get_svc_primary_ips_param, make_response
+)
+from .constants import FAMILY_PARAM_NAME, AS_CIDR_PARAM_NAME
 
 
 FAMILY_PARAM = openapi.Parameter(
-    "family",
+    FAMILY_PARAM_NAME,
     in_=openapi.IN_QUERY,
     description="Filter IPs and or prefixes by address family (4|6).",
     type=openapi.TYPE_INTEGER,
     enum=[4, 6]
 )
 AS_CIDR_PARAM = openapi.Parameter(
-    "as_cidr", in_=openapi.IN_QUERY,
+    AS_CIDR_PARAM_NAME, in_=openapi.IN_QUERY,
     description="Return IPs as /32 or /128", type=openapi.TYPE_BOOLEAN
 )
 
@@ -43,10 +50,25 @@ class ListsRootView(APIRootView):
         return "Lists"
 
 
-class ValuesListViewSet(GenericViewSet):
+class ListsBaseViewSet(GenericViewSet):
     renderer_classes = [JSONRenderer, BrowsableAPIRenderer, PlainTextRenderer]
 
-    def list(self, request):
+    # Adapted from
+    # https://github.com/netbox-community/netbox/blob/a33e47780b42f49f4ea536bace1617fa7dda31ab/
+    # netbox/netbox/api/views.py#L179
+    def initial(self, request, *args, **kwargs):
+        super().initial(request, *args, **kwargs)
+
+        if not request.user.is_authenticated:
+            return
+
+        # Restrict the view's QuerySet to allow only the permitted objects
+        self.queryset = self.queryset.restrict(request.user, "view")
+
+
+class ValuesListViewSet(ListsBaseViewSet):
+
+    def list(self, request: Request) -> Response:
         queryset = self.filter_queryset(self.get_queryset())
 
         return Response([str(i) for i in queryset])
@@ -54,7 +76,7 @@ class ValuesListViewSet(GenericViewSet):
 
 class PrefixListViewSet(ValuesListViewSet):
     queryset = Prefix.objects.values_list("prefix", flat=True).distinct()
-    filterset_class = PrefixFilterSet
+    filterset_class = CustomPrefixFilterSet
 
 
 class AggregateListViewSet(ValuesListViewSet):
@@ -70,28 +92,29 @@ class IPAddressListViewSet(ValuesListViewSet):
     def list(self, request):
         queryset = self.filter_queryset(self.get_queryset())
 
-        if "as_cidr" in request.query_params:
-            return Response(list(set([as_cidr(i) for i in queryset])))
-        # We use list/set because distinct() won't work
-        # for two IPs with the same address but different prefix length.
-        return Response(list(set([str(i.ip) for i in queryset])))
+        return make_response([format_ipn(i, get_as_cidr(request)) for i in queryset])
 
 
 class ServiceListviewSet(ValuesListViewSet):
-    queryset = Service.objects.filter(ipaddresses__isnull=False).values_list(
-        "ipaddresses__address", flat=True
-    ).distinct()
+    queryset = Service.objects.all()
 
     filterset_class = ServiceFilterSet
 
-    @swagger_auto_schema(manual_parameters=[AS_CIDR_PARAM])
-    def list(self, request):
-        queryset = self.filter_queryset(self.get_queryset())
+    @swagger_auto_schema(manual_parameters=[
+        AS_CIDR_PARAM,
+        openapi.Parameter(
+            "primary_ips", in_=openapi.IN_QUERY,
+            description="Return Primary IPs if the service doesn't have any assigned IPs.",
+            type=openapi.TYPE_BOOLEAN
+        )
+    ])
+    def list(self, request: Request) -> Response:
+        as_cidr = get_as_cidr(request)
+        family = get_family_param(request)
+        primary_ips = get_svc_primary_ips_param("primary_ips", request)
 
-        if "as_cidr" in request.query_params:
-            return Response(list(set([as_cidr(i) for i in queryset])))
-
-        return Response(list(set([str(i.ip) for i in queryset])))
+        qs = self.filter_queryset(self.get_queryset())
+        return make_response(get_service_ips(qs, as_cidr, family, primary_ips))
 
 
 class DevicesListViewSet(ValuesListViewSet):
@@ -102,43 +125,168 @@ class DevicesListViewSet(ValuesListViewSet):
         operation_description="Returns the primary IPs of devices.",
         manual_parameters=[AS_CIDR_PARAM, FAMILY_PARAM]
     )
-    def list(self, request: Request):
+    def list(self, request: Request) -> Response:
 
-        family = request.query_params.get("family", None)
-        validateFamilyParam(family)
+        family = get_family_param(request)
 
-        return Response(devicesVmPrimaryList(
+        return Response(device_vm_primary_list(
             self.filter_queryset((self.get_queryset())),
             family,
-            cidr=("as_cidr" in request.query_params)
+            cidr=get_as_cidr(request)
         ))
 
 
-@swagger_auto_schema(operation_description="Returns the primary IPs of devices.")
 class VirtualMachinesListViewSet(ValuesListViewSet):
     queryset = VirtualMachine.objects.all()
     filterset_class = VirtualMachineFilterSet
 
     @swagger_auto_schema(
-        operation_description="Returns the primary IPs of devices.",
+        operation_description="Returns the primary IPs of virtual machines.",
         manual_parameters=[AS_CIDR_PARAM, FAMILY_PARAM]
     )
-    def list(self, request: Request):
-        family = request.query_params.get("family", None)
-        validateFamilyParam(family)
+    def list(self, request: Request) -> Response:
+        family = get_family_param(request)
 
-        return Response(devicesVmPrimaryList(
+        return Response(device_vm_primary_list(
             self.filter_queryset((self.get_queryset())),
             family,
-            cidr=("as_cidr" in request.query_params)
+            cidr=get_as_cidr(request)
         ))
 
 
-class TagsListViewSet(GenericViewSet):
+class DevicesVMsListView(APIView):
+    renderer_classes = [JSONRenderer, BrowsableAPIRenderer, PlainTextRenderer]
+    # We need to have `queryset defined`. Otherwise, the following occurs:
+    # Cannot apply TokenPermissions on a view that does not set `.queryset` or have a `.get_queryset()` method.
+    # See https://github.com/encode/django-rest-framework/blob/
+    # 71e6c30034a1dd35a39ca74f86c371713e762c79/rest_framework/permissions.py#L207
+    #
+    # Therefore, we use Device as the model.
+    queryset = Device.objects.all()
+
+    @swagger_auto_schema(
+        operation_description="Combined devices and virtual machines primary IPs list. "
+        "Use only parameters common to both devices and VMs."
+    )
+    def get(self, request: Request) -> Response:
+        family = get_family_param(request)
+        as_cidr = get_as_cidr(request)
+        devices_fs = DeviceFilterSet(request.query_params, queryset=Device.objects.restrict(request.user, "view").all())
+        vms_fs = VirtualMachineFilterSet(
+            request.query_params,
+            queryset=VirtualMachine.objects.restrict(request.user, "view").all()
+        )
+        devices = device_vm_primary_list(devices_fs.qs, family, as_cidr)
+        vms = device_vm_primary_list(vms_fs.qs, family, as_cidr)
+        return make_response(devices + vms)
+
+
+class TagsListViewSet(ListsBaseViewSet):
     queryset = Tag.objects.all()
     lookup_field = "slug"
     lookup_value_regex = r"[-\w]+"
-    renderer_classes = [JSONRenderer, BrowsableAPIRenderer, PlainTextRenderer]
+
+    def param_all_any(self, request: Request, param: str) -> bool:
+        return (
+            param in request.query_params
+            or "all" in request.query_params
+            or "all_primary" in request.query_params
+        )
+
+    def param_all_primary(self, request: Request, param: str, primary: bool) -> bool:
+        return (
+            param in request.query_params
+            or ("all" in request.query_params and not primary)
+            or ("all_primary" in request.query_params and primary)
+        )
+
+    def get_prefixes(self, tag: Tag, family: Union[int, None], request: Request) -> List[str]:
+        if not self.param_all_any(request, "prefixes"):
+            return []
+
+        if family == 4:
+            family_filter = Q(prefix__family=4)
+        elif family == 6:
+            family_filter = Q(prefix__family=6)
+        else:
+            family_filter = Q()
+        qs = Prefix.objects.restrict(request.user, "view").filter(
+            Q(tags=tag) & family_filter
+        ).values_list("prefix", flat=True).distinct()
+        return [str(i) for i in qs]
+
+    def get_aggregates(self, tag: Tag, family: Union[int, None], request: Request) -> List[str]:
+        if not self.param_all_any(request, "aggregates"):
+            return []
+
+        if family == 4:
+            family_filter = Q(prefix__family=4)
+        elif family == 6:
+            family_filter = Q(prefix__family=6)
+        else:
+            family_filter = Q()
+
+        qs = Aggregate.objects.restrict(request.user, "view").filter(
+            Q(tags=tag) & family_filter
+        ).values_list("prefix", flat=True).distinct()
+        return [str(i) for i in qs]
+
+    def get_ips(self, tag: Tag, family: Union[int, None], request: Request) -> List[str]:
+        if family == 4:
+            family_filter = Q(address__family=4)
+        elif family == 6:
+            family_filter = Q(address__family=6)
+        else:
+            family_filter = Q()
+
+        ip_filters = []
+        if self.param_all_any(request, "ips"):
+            ip_filters.append(Q(tags=tag))
+        if self.param_all_primary(request, "devices", False):
+            ip_filters.append(Q(interface__device__tags=tag))
+        if self.param_all_primary(request, "vms", False):
+            ip_filters.append(Q(vminterface__virtual_machine__tags=tag))
+
+        if len(ip_filters) > 0:
+            return [
+                format_ipn(i, True)
+                for i in IPAddress.objects.restrict(request.user, "view").filter(
+                    reduce(operator.or_, ip_filters) & family_filter
+                ).values_list("address", flat=True).distinct()
+            ]
+        else:
+            return []
+
+    def get_services(self, tag: Tag, family: Union[int, None], request: Request) -> List[str]:
+        if not self.param_all_any(request, "services"):
+            return []
+
+        return get_service_ips(
+            Service.objects.restrict(request.user, "view").filter(tags=tag),
+            True,
+            family,
+            get_svc_primary_ips_param("service_primary_ips", request)
+        )
+
+    def get_devices_primary(self, tag: Tag, family: Union[int, None], request: Request) -> List[str]:
+        if not self.param_all_primary(request, "devices_primary", True):
+            return []
+
+        return device_vm_primary_list(
+            Device.objects.restrict(request.user, "view").filter(tags=tag),
+            family,
+            cidr=True
+        )
+
+    def get_vms_primary(self, tag: Tag, family: Union[int, None], request: Request) -> List[str]:
+        if not self.param_all_primary(request, "vms_primary", True):
+            return []
+
+        return device_vm_primary_list(
+            VirtualMachine.objects.restrict(request.user, "view").filter(tags=tag),
+            family,
+            cidr=True
+        )
 
     @swagger_auto_schema(manual_parameters=[
         openapi.Parameter(
@@ -172,91 +320,34 @@ class TagsListViewSet(GenericViewSet):
         openapi.Parameter(
             "ips", in_=openapi.IN_QUERY,
             description="Include IP Addresses", type=openapi.TYPE_BOOLEAN
+        ),
+        openapi.Parameter(
+            "service_primary_ips", in_=openapi.IN_QUERY,
+            description="Return Primary IPs if the service doesn't have any assigned IPs.", type=openapi.TYPE_BOOLEAN
+        ),
+        openapi.Parameter(
+            "all", in_=openapi.IN_QUERY,
+            description="Include ALL of the above options except *_primary.", type=openapi.TYPE_BOOLEAN
+        ),
+        openapi.Parameter(
+            "all_primary", in_=openapi.IN_QUERY,
+            description="Include ALL of the above options, using Device/VM primary IPs.", type=openapi.TYPE_BOOLEAN
         )
     ])
-    def retrieve(self, request, slug=None):
+    def retrieve(self, request, slug=None) -> Response:
         if not slug:
             return Response("No slug", status.HTTP_400_BAD_REQUEST)
 
         tag = get_object_or_404(Tag, slug=slug)
+        family = get_family_param(request)
+        prefixes = self.get_prefixes(tag, family, request)
+        aggregates = self.get_aggregates(tag, family, request)
+        ips = self.get_ips(tag, family, request)
+        services = self.get_services(tag, family, request)
+        devices_primary = self.get_devices_primary(tag, family, request)
+        vms_primary = self.get_vms_primary(tag, family, request)
 
-        family = request.query_params.get("family", None)
-        validateFamilyParam(family)
-
-        if family == '4':
-            family_filter = Q(family=4)
-        elif family == '6':
-            family_filter = Q(family=6)
-        else:
-            family_filter = Q()
-
-        if "prefixes" in request.query_params:
-            prefixes = [
-                str(i) for i in Prefix.objects.filter(tags=tag).values_list("prefix", flat=True).distinct()
-            ]
-        else:
-            prefixes = []
-
-        if "aggregates" in request.query_params:
-            aggregates = [
-                str(i) for i in Aggregate.objects.filter(tags=tag).values_list("prefix", flat=True).distinct()
-            ]
-        else:
-            aggregates = []
-
-        ip_filters = []
-        if "ips" in request.query_params:
-            ip_filters.append(Q(tags=tag))
-        if "devices" in request.query_params:
-            ip_filters.append(Q(interface__device__tags=tag))
-        if "vms" in request.query_params:
-            ip_filters.append(Q(vminterface__virtual_machine__tags=tag))
-        if len(ip_filters) > 0:
-            ips = [
-                as_cidr(i)
-                for i in IPAddress.objects.filter(
-                    reduce(operator.or_, ip_filters) & family_filter
-                ).values_list("address", flat=True).distinct()
-            ]
-        else:
-            ips = []
-
-        if "services" in request.query_params:
-            if family == ' 4':
-                svc_family_filter = Q(ipaddresses__address__family=4)
-            elif family == '6':
-                svc_family_filter = Q(ipaddresses__address__family=6)
-            else:
-                svc_family_filter = Q()
-
-            services = [
-                as_cidr(i)
-                for i in Service.objects.filter(
-                    Q(ipaddresses__isnull=False) & Q(tags=tag) & svc_family_filter
-                ).values_list("ipaddresses__address", flat=True).distinct()
-            ]
-        else:
-            services = []
-
-        if "devices_primary" in request.query_params:
-            devices_primary = devicesVmPrimaryList(
-                Device.objects.filter(tags=tag),
-                family,
-                cidr=True
-            )
-        else:
-            devices_primary = []
-
-        if "vms_primary" in request.query_params:
-            vms_primary = devicesVmPrimaryList(
-                VirtualMachine.objects.filter(tags=tag),
-                family,
-                cidr=True
-            )
-        else:
-            vms_primary = []
-
-        return Response(list(set(prefixes + aggregates + ips + services + devices_primary + vms_primary)))
+        return make_response(prefixes + aggregates + ips + services + devices_primary + vms_primary)
 
 
 class PrometheusDeviceSD(GenericViewSet):
@@ -283,7 +374,7 @@ class PrometheusDeviceSD(GenericViewSet):
             "labels": labels
         }
 
-    def list(self, request):
+    def list(self, request: Request) -> Response:
         queryset = self.filter_queryset(self.get_queryset())
         return Response([self._sd_device(d) for d in queryset])
 
@@ -294,7 +385,7 @@ class PrometheusVirtualMachineSD(GenericViewSet):
 
     def _sd_vm(self, vm: VirtualMachine) -> dict:
         labels = {
-            "__meta_netbox_id": vm.id,
+            "__meta_netbox_id": str(vm.id),
             "__meta_netbox_name": vm.name,
             "__meta_netbox_status": vm.status,
             "__meta_netbox_cluster_name": vm.cluster.name,
@@ -306,13 +397,13 @@ class PrometheusVirtualMachineSD(GenericViewSet):
             "__meta_netbox_primary_ip6": str(vm.primary_ip6.address.ip) if vm.primary_ip6 else ""
         }
         for k, v in vm.custom_field_data.items():
-            labels[f"__meta_netbox_cf_{k}"] = v
+            labels[f"__meta_netbox_cf_{k}"] = str(v)
 
         return {
             "targets": [str(vm.primary_ip.address.ip) if vm.primary_ip else vm.name],
             "labels": labels
         }
 
-    def list(self, request):
+    def list(self, request: Request) -> Response:
         queryset = self.filter_queryset(self.get_queryset())
         return Response([self._sd_vm(vm) for vm in queryset])
